@@ -10,7 +10,6 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import builtins
 import datetime
 from http import client as http_client
 import io
@@ -43,7 +42,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.service.validate_href(self.href)
         path_mock.assert_not_called()
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         response.status_code = http_client.NO_CONTENT
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href,
@@ -61,7 +60,7 @@ class HttpImageServiceTestCase(base.TestCase):
         response.status_code = http_client.OK
         self.service.validate_href(self.href)
         head_mock.assert_called_once_with(self.href, verify=False,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         response.status_code = http_client.NO_CONTENT
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href,
@@ -78,7 +77,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         head_mock.assert_called_once_with(self.href, verify=False,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         head_mock.side_effect = requests.RequestException()
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
@@ -91,7 +90,7 @@ class HttpImageServiceTestCase(base.TestCase):
         response.status_code = http_client.OK
         self.service.validate_href(self.href)
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         response.status_code = http_client.NO_CONTENT
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href,
@@ -109,7 +108,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         head_mock.side_effect = requests.RequestException()
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
@@ -123,7 +122,7 @@ class HttpImageServiceTestCase(base.TestCase):
 
         self.service.validate_href(self.href)
         head_mock.assert_called_once_with(self.href, verify='/some/path',
-                                          timeout=60)
+                                          timeout=60, auth=None)
         response.status_code = http_client.NO_CONTENT
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href,
@@ -133,6 +132,43 @@ class HttpImageServiceTestCase(base.TestCase):
                           self.service.validate_href,
                           self.href)
 
+    @mock.patch.object(requests, 'head', autospec=True)
+    def test_validate_href_valid_path_valid_basic_auth(self, head_mock):
+        cfg.CONF.set_override('webserver_verify_ca', '/some/path')
+        cfg.CONF.set_override('image_server_auth_strategy',
+                              'http_basic',
+                              'deploy')
+        cfg.CONF.set_override('image_server_user', 'test', 'deploy')
+        cfg.CONF.set_override('image_server_password', 'test', 'deploy')
+        user = cfg.CONF.deploy.image_server_user
+        password = cfg.CONF.deploy.image_server_password
+        auth_creds = requests.auth.HTTPBasicAuth(user, password)
+        response = head_mock.return_value
+        response.status_code = http_client.OK
+
+        self.service.validate_href(self.href)
+        head_mock.assert_called_once_with(self.href, verify='/some/path',
+                                          timeout=60, auth=auth_creds)
+        response.status_code = http_client.NO_CONTENT
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.validate_href,
+                          self.href)
+        response.status_code = http_client.BAD_REQUEST
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.validate_href,
+                          self.href)
+
+    @mock.patch.object(requests, 'head', autospec=True)
+    def test_validate_href_valid_path_invalid_basic_auth(self, head_mock):
+        cfg.CONF.set_override('webserver_verify_ca', '/some/path')
+        cfg.CONF.set_override('image_server_auth_strategy',
+                              'http_basic',
+                              'deploy')
+
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.validate_href,
+                          self.href)
+
     @mock.patch.object(requests, 'head', autospec=True)
     def test_validate_href_custom_timeout(self, head_mock):
         cfg.CONF.set_override('webserver_connection_timeout', 15)
@@ -141,7 +177,7 @@ class HttpImageServiceTestCase(base.TestCase):
         response.status_code = http_client.OK
         self.service.validate_href(self.href)
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=15)
+                                          timeout=15, auth=None)
         response.status_code = http_client.NO_CONTENT
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href,
@@ -161,7 +197,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         head_mock.assert_called_once_with(self.href, verify='/some/path',
-                                          timeout=60)
+                                          timeout=60, auth=None)
 
     @mock.patch.object(requests, 'head', autospec=True)
     def test_validate_href_verify_error(self, head_mock):
@@ -170,7 +206,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         head_mock.assert_called_once_with(self.href, verify='/some/path',
-                                          timeout=60)
+                                          timeout=60, auth=None)
 
     @mock.patch.object(requests, 'head', autospec=True)
     def test_validate_href_verify_os_error(self, head_mock):
@@ -179,7 +215,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         head_mock.assert_called_once_with(self.href, verify='/some/path',
-                                          timeout=60)
+                                          timeout=60, auth=None)
 
     @mock.patch.object(requests, 'head', autospec=True)
     def test_validate_href_error_with_secret_parameter(self, head_mock):
@@ -192,7 +228,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertIn('secreturl', str(e))
         self.assertNotIn(self.href, str(e))
         head_mock.assert_called_once_with(self.href, verify=False,
-                                          timeout=60)
+                                          timeout=60, auth=None)
 
     @mock.patch.object(requests, 'head', autospec=True)
     def test_validate_href_path_forbidden(self, head_mock):
@@ -203,7 +239,7 @@ class HttpImageServiceTestCase(base.TestCase):
         url = self.href + '/'
         resp = self.service.validate_href(url)
         head_mock.assert_called_once_with(url, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         self.assertEqual(http_client.FORBIDDEN, resp.status_code)
 
     @mock.patch.object(requests, 'head', autospec=True)
@@ -220,7 +256,63 @@ class HttpImageServiceTestCase(base.TestCase):
                                 url)
         self.assertEqual(new_url, exc.redirect_url)
         head_mock.assert_called_once_with(url, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
+
+    def test_verify_basic_auth_cred_format(self):
+        self.assertIsNone(self
+                          .service
+                          .verify_basic_auth_cred_format(self.href,
+                                                         "SpongeBob",
+                                                         "SquarePants"))
+
+    def test_verify_basic_auth_cred_format_empty_user(self):
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.verify_basic_auth_cred_format,
+                          self.href,
+                          "",
+                          "SquarePants")
+
+    def test_verify_basic_auth_cred_format_empty_password(self):
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.verify_basic_auth_cred_format,
+                          self.href,
+                          "SpongeBob",
+                          "")
+
+    def test_verify_basic_auth_cred_format_none_user(self):
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.verify_basic_auth_cred_format,
+                          self.href,
+                          None,
+                          "SquarePants")
+
+    def test_verify_basic_auth_cred_format_none_password(self):
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.verify_basic_auth_cred_format,
+                          self.href,
+                          "SpongeBob",
+                          None)
+
+    def test_gen_auth_from_conf_user_pass_success(self):
+        cfg.CONF.set_override('image_server_auth_strategy',
+                              'http_basic',
+                              'deploy')
+        cfg.CONF.set_override('image_server_password', 'SpongeBob', 'deploy')
+        cfg.CONF.set_override('image_server_user', 'SquarePants', 'deploy')
+        correct_auth = \
+            requests.auth.HTTPBasicAuth('SquarePants',
+                                        'SpongeBob')
+        return_auth = \
+            self.service.gen_auth_from_conf_user_pass(self.href)
+        self.assertEqual(correct_auth, return_auth)
+
+    def test_gen_auth_from_conf_user_pass_none(self):
+        cfg.CONF.set_override('image_server_auth_strategy', 'noauth', 'deploy')
+        cfg.CONF.set_override('image_server_password', 'SpongeBob', 'deploy')
+        cfg.CONF.set_override('image_server_user', 'SquarePants', 'deploy')
+        return_auth = \
+            self.service.gen_auth_from_conf_user_pass(self.href)
+        self.assertIsNone(return_auth)
 
     @mock.patch.object(requests, 'head', autospec=True)
     def _test_show(self, head_mock, mtime, mtime_date):
@@ -231,7 +323,7 @@ class HttpImageServiceTestCase(base.TestCase):
         }
         result = self.service.show(self.href)
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         self.assertEqual({'size': 100, 'updated_at': mtime_date,
                           'properties': {}, 'no_cache': False}, result)
 
@@ -257,7 +349,7 @@ class HttpImageServiceTestCase(base.TestCase):
         }
         result = self.service.show(self.href)
         head_mock.assert_called_once_with(self.href, verify=True,
-                                          timeout=60)
+                                          timeout=60, auth=None)
         self.assertEqual({
             'size': 100,
             'updated_at': datetime.datetime(2014, 11, 15, 8, 12, 31),
@@ -281,7 +373,7 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.show, self.href)
         head_mock.assert_called_with(self.href, verify=True,
-                                     timeout=60)
+                                     timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -298,7 +390,7 @@ class HttpImageServiceTestCase(base.TestCase):
         )
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify=True,
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -316,7 +408,42 @@ class HttpImageServiceTestCase(base.TestCase):
         )
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify=False,
-                                             timeout=60)
+                                             timeout=60, auth=None)
+
+    @mock.patch.object(shutil, 'copyfileobj', autospec=True)
+    @mock.patch.object(requests, 'get', autospec=True)
+    def test_download_success_verify_false_basic_auth_sucess(
+            self, req_get_mock, shutil_mock):
+        cfg.CONF.set_override('webserver_verify_ca', 'False')
+        cfg.CONF.set_override('image_server_auth_strategy',
+                              'http_basic',
+                              'deploy')
+        cfg.CONF.set_override('image_server_user', 'test', 'deploy')
+        cfg.CONF.set_override('image_server_password', 'test', 'deploy')
+        user = cfg.CONF.deploy.image_server_user
+        password = cfg.CONF.deploy.image_server_password
+        auth_creds = requests.auth.HTTPBasicAuth(user, password)
+        response_mock = req_get_mock.return_value
+        response_mock.status_code = http_client.OK
+        response_mock.raw = mock.MagicMock(spec=io.BytesIO)
+        file_mock = mock.Mock(spec=io.BytesIO)
+        self.service.download(self.href, file_mock)
+        shutil_mock.assert_called_once_with(
+            response_mock.raw.__enter__(), file_mock,
+            image_service.IMAGE_CHUNK_SIZE
+        )
+        req_get_mock.assert_called_once_with(self.href, stream=True,
+                                             verify=False, timeout=60,
+                                             auth=auth_creds)
+
+    def test_download_success_verify_false_basic_auth_failed(self):
+        cfg.CONF.set_override('webserver_verify_ca', 'False')
+        cfg.CONF.set_override('image_server_auth_strategy',
+                              'http_basic',
+                              'deploy')
+        file_mock = mock.Mock(spec=io.BytesIO)
+        self.assertRaises(exception.ImageRefValidationFailed,
+                          self.service.download, self.href, file_mock)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -334,7 +461,7 @@ class HttpImageServiceTestCase(base.TestCase):
         )
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify=True,
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -352,7 +479,7 @@ class HttpImageServiceTestCase(base.TestCase):
         )
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify='/some/path',
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -377,8 +504,8 @@ class HttpImageServiceTestCase(base.TestCase):
         self.assertRaises(exception.ImageDownloadFailed,
                           self.service.download, self.href, file_mock)
         req_get_mock.assert_called_once_with(self.href, stream=True,
-                                             verify=False,
-                                             timeout=60)
+                                             verify=False, timeout=60,
+                                             auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -395,7 +522,7 @@ class HttpImageServiceTestCase(base.TestCase):
                           self.service.download, self.href, file_mock)
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify='/some/path',
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -411,7 +538,7 @@ class HttpImageServiceTestCase(base.TestCase):
                           self.service.download, self.href, file_mock)
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify='/some/path',
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -427,7 +554,7 @@ class HttpImageServiceTestCase(base.TestCase):
                           self.service.download, self.href, file_mock)
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify='/some/path',
-                                             timeout=60)
+                                             timeout=60, auth=None)
 
     @mock.patch.object(shutil, 'copyfileobj', autospec=True)
     @mock.patch.object(requests, 'get', autospec=True)
@@ -445,27 +572,99 @@ class HttpImageServiceTestCase(base.TestCase):
         )
         req_get_mock.assert_called_once_with(self.href, stream=True,
                                              verify=True,
-                                             timeout=15)
+                                             timeout=15, auth=None)
+
+    @mock.patch.object(requests, 'get', autospec=True)
+    def test_get_success(self, req_get_mock):
+        response_mock = req_get_mock.return_value
+        response_mock.status_code = http_client.OK
+        response_mock.text = 'value'
+        self.assertEqual('value', self.service.get('http://url'))
+        req_get_mock.assert_called_once_with('http://url', stream=False,
+                                             verify=True, timeout=60,
+                                             auth=None)
+
+    @mock.patch.object(requests, 'get', autospec=True)
+    def test_get_handles_exceptions(self, req_get_mock):
+        for exc in [OSError, requests.ConnectionError,
+                    requests.RequestException, IOError]:
+            req_get_mock.reset_mock()
+            req_get_mock.side_effect = exc
+            self.assertRaises(exception.ImageDownloadFailed,
+                              self.service.get,
+                              'http://url')
+            req_get_mock.assert_called_once_with('http://url', stream=False,
+                                                 verify=True, timeout=60,
+                                                 auth=None)
+
+    @mock.patch.object(requests, 'get', autospec=True)
+    def test_get_success_verify_false(self, req_get_mock):
+        cfg.CONF.set_override('webserver_verify_ca', False)
+        response_mock = req_get_mock.return_value
+        response_mock.status_code = http_client.OK
+        response_mock.text = 'value'
+        self.assertEqual('value', self.service.get('http://url'))
+        req_get_mock.assert_called_once_with('http://url', stream=False,
+                                             verify=False, timeout=60,
+                                             auth=None)
 
 
 class FileImageServiceTestCase(base.TestCase):
     def setUp(self):
         super(FileImageServiceTestCase, self).setUp()
         self.service = image_service.FileImageService()
-        self.href = 'file:///home/user/image.qcow2'
-        self.href_path = '/home/user/image.qcow2'
+        self.href = 'file:///var/lib/ironic/images/image.qcow2'
+        self.href_path = '/var/lib/ironic/images/image.qcow2'
 
     @mock.patch.object(os.path, 'isfile', return_value=True, autospec=True)
     def test_validate_href(self, path_exists_mock):
         self.service.validate_href(self.href)
         path_exists_mock.assert_called_once_with(self.href_path)
 
-    @mock.patch.object(os.path, 'isfile', return_value=False, autospec=True)
+    @mock.patch.object(os.path, 'isfile', return_value=False,
+                       autospec=True)
     def test_validate_href_path_not_found_or_not_file(self, path_exists_mock):
         self.assertRaises(exception.ImageRefValidationFailed,
                           self.service.validate_href, self.href)
         path_exists_mock.assert_called_once_with(self.href_path)
 
+    @mock.patch.object(os.path, 'abspath', autospec=True)
+    def test_validate_href_empty_allowlist(self, abspath_mock):
+        abspath_mock.return_value = self.href_path
+        cfg.CONF.set_override('file_url_allowed_paths', [], 'conductor')
+        self.assertRaisesRegex(exception.ImageRefValidationFailed,
+                               "is not allowed for image source file URLs",
+                               self.service.validate_href, self.href)
+
+    @mock.patch.object(os.path, 'abspath', autospec=True)
+    def test_validate_href_not_in_allowlist(self, abspath_mock):
+        href = "file:///var/is/allowed/not/this/path/image.qcow2"
+        href_path = "/var/is/allowed/not/this/path/image.qcow2"
+        abspath_mock.side_effect = ['/var/lib/ironic', href_path]
+        cfg.CONF.set_override('file_url_allowed_paths', ['/var/lib/ironic'],
+                              'conductor')
+        self.assertRaisesRegex(exception.ImageRefValidationFailed,
+                               "is not allowed for image source file URLs",
+                               self.service.validate_href, href)
+
+    @mock.patch.object(os.path, 'abspath', autospec=True)
+    @mock.patch.object(os.path, 'isfile',
+                       return_value=True, autospec=True)
+    def test_validate_href_in_allowlist(self,
+                                        path_exists_mock,
+                                        abspath_mock):
+        href_dir = '/var/lib'  # self.href_path is in /var/lib/ironic/images/
+        # First call is ironic.conf.types.ExplicitAbsolutePath
+        # Second call is in validate_href()
+        abspath_mock.side_effect = [href_dir, self.href_path]
+        cfg.CONF.set_override('file_url_allowed_paths', [href_dir],
+                              'conductor')
+        result = self.service.validate_href(self.href)
+        self.assertEqual(self.href_path, result)
+        path_exists_mock.assert_called_once_with(self.href_path)
+        abspath_mock.assert_has_calls(
+            [mock.call(href_dir), mock.call(self.href_path)])
+
     @mock.patch.object(os.path, 'getmtime', return_value=1431087909.1641912,
                        autospec=True)
     @mock.patch.object(os.path, 'getsize', return_value=42, autospec=True)
@@ -483,119 +682,74 @@ class FileImageServiceTestCase(base.TestCase):
                           'properties': {},
                           'no_cache': True}, result)
 
+    @mock.patch.object(shutil, 'copyfile', autospec=True)
     @mock.patch.object(os, 'link', autospec=True)
+    @mock.patch.object(os.path, 'realpath', lambda p: p)
     @mock.patch.object(os, 'remove', autospec=True)
-    @mock.patch.object(os, 'access', return_value=True, autospec=True)
-    @mock.patch.object(os, 'stat', autospec=True)
     @mock.patch.object(image_service.FileImageService, 'validate_href',
                        autospec=True)
-    def test_download_hard_link(self, _validate_mock, stat_mock, access_mock,
-                                remove_mock, link_mock):
+    def test_download_hard_link(self, _validate_mock, remove_mock, link_mock,
+                                copy_mock):
         _validate_mock.return_value = self.href_path
-        stat_mock.return_value.st_dev = 'dev1'
         file_mock = mock.Mock(spec=io.BytesIO)
         file_mock.name = 'file'
         self.service.download(self.href, file_mock)
         _validate_mock.assert_called_once_with(mock.ANY, self.href)
-        self.assertEqual(2, stat_mock.call_count)
-        access_mock.assert_called_once_with(self.href_path, os.R_OK | os.W_OK)
         remove_mock.assert_called_once_with('file')
         link_mock.assert_called_once_with(self.href_path, 'file')
+        copy_mock.assert_not_called()
 
-    @mock.patch.object(os, 'sendfile', return_value=42, autospec=True)
-    @mock.patch.object(os.path, 'getsize', return_value=42, autospec=True)
-    @mock.patch.object(builtins, 'open', autospec=True)
-    @mock.patch.object(os, 'access', return_value=False, autospec=True)
-    @mock.patch.object(os, 'stat', autospec=True)
+    @mock.patch.object(shutil, 'copyfile', autospec=True)
+    @mock.patch.object(os, 'link', autospec=True)
+    @mock.patch.object(os, 'remove', autospec=True)
     @mock.patch.object(image_service.FileImageService, 'validate_href',
                        autospec=True)
-    def test_download_copy(self, _validate_mock, stat_mock, access_mock,
-                           open_mock, size_mock, copy_mock):
+    def test_download_copy(self, _validate_mock, remove_mock, link_mock,
+                           copy_mock):
         _validate_mock.return_value = self.href_path
-        stat_mock.return_value.st_dev = 'dev1'
+        link_mock.side_effect = PermissionError
         file_mock = mock.MagicMock(spec=io.BytesIO)
         file_mock.name = 'file'
-        input_mock = mock.MagicMock(spec=io.BytesIO)
-        open_mock.return_value = input_mock
         self.service.download(self.href, file_mock)
         _validate_mock.assert_called_once_with(mock.ANY, self.href)
-        self.assertEqual(2, stat_mock.call_count)
-        access_mock.assert_called_once_with(self.href_path, os.R_OK | os.W_OK)
-        copy_mock.assert_called_once_with(file_mock.fileno(),
-                                          input_mock.__enter__().fileno(),
-                                          0, 42)
+        link_mock.assert_called_once_with(self.href_path, 'file')
+        copy_mock.assert_called_once_with(self.href_path, 'file')
 
-    @mock.patch.object(os, 'sendfile', autospec=True)
-    @mock.patch.object(os.path, 'getsize', return_value=42, autospec=True)
-    @mock.patch.object(builtins, 'open', autospec=True)
-    @mock.patch.object(os, 'access', return_value=False, autospec=True)
-    @mock.patch.object(os, 'stat', autospec=True)
+    @mock.patch.object(shutil, 'copyfile', autospec=True)
+    @mock.patch.object(os, 'link', autospec=True)
+    @mock.patch.object(os.path, 'realpath', autospec=True)
+    @mock.patch.object(os, 'remove', autospec=True)
     @mock.patch.object(image_service.FileImageService, 'validate_href',
                        autospec=True)
-    def test_download_copy_segmented(self, _validate_mock, stat_mock,
-                                     access_mock, open_mock, size_mock,
-                                     copy_mock):
-        # Fake a 3G + 1k image
-        chunk_size = image_service.SENDFILE_CHUNK_SIZE
-        fake_image_size = chunk_size * 3 + 1024
-        fake_chunk_seq = [chunk_size, chunk_size, chunk_size, 1024]
+    def test_download_symlink(self, _validate_mock, remove_mock,
+                              realpath_mock, link_mock, copy_mock):
         _validate_mock.return_value = self.href_path
-        stat_mock.return_value.st_dev = 'dev1'
+        realpath_mock.side_effect = lambda p: p + '.real'
         file_mock = mock.MagicMock(spec=io.BytesIO)
         file_mock.name = 'file'
-        input_mock = mock.MagicMock(spec=io.BytesIO)
-        open_mock.return_value = input_mock
-        size_mock.return_value = fake_image_size
-        copy_mock.side_effect = fake_chunk_seq
         self.service.download(self.href, file_mock)
         _validate_mock.assert_called_once_with(mock.ANY, self.href)
-        self.assertEqual(2, stat_mock.call_count)
-        access_mock.assert_called_once_with(self.href_path, os.R_OK | os.W_OK)
-        copy_calls = [mock.call(file_mock.fileno(),
-                                input_mock.__enter__().fileno(),
-                                chunk_size * i,
-                                fake_chunk_seq[i]) for i in range(4)]
-        copy_mock.assert_has_calls(copy_calls)
-        size_mock.assert_called_once_with(self.href_path)
+        realpath_mock.assert_called_once_with(self.href_path)
+        link_mock.assert_called_once_with(self.href_path + '.real', 'file')
+        copy_mock.assert_not_called()
 
-    @mock.patch.object(os, 'remove', side_effect=OSError, autospec=True)
-    @mock.patch.object(os, 'access', return_value=True, autospec=True)
-    @mock.patch.object(os, 'stat', autospec=True)
+    @mock.patch.object(shutil, 'copyfile', autospec=True)
+    @mock.patch.object(os, 'link', autospec=True)
+    @mock.patch.object(os, 'remove', autospec=True)
     @mock.patch.object(image_service.FileImageService, 'validate_href',
                        autospec=True)
-    def test_download_hard_link_fail(self, _validate_mock, stat_mock,
-                                     access_mock, remove_mock):
+    def test_download_copy_fail(self, _validate_mock, remove_mock, link_mock,
+                                copy_mock):
         _validate_mock.return_value = self.href_path
-        stat_mock.return_value.st_dev = 'dev1'
+        link_mock.side_effect = PermissionError
+        copy_mock.side_effect = PermissionError
         file_mock = mock.MagicMock(spec=io.BytesIO)
         file_mock.name = 'file'
         self.assertRaises(exception.ImageDownloadFailed,
                           self.service.download, self.href, file_mock)
         _validate_mock.assert_called_once_with(mock.ANY, self.href)
-        self.assertEqual(2, stat_mock.call_count)
-        access_mock.assert_called_once_with(self.href_path, os.R_OK | os.W_OK)
-
-    @mock.patch.object(os, 'sendfile', side_effect=OSError, autospec=True)
-    @mock.patch.object(os.path, 'getsize', return_value=42, autospec=True)
-    @mock.patch.object(builtins, 'open', autospec=True)
-    @mock.patch.object(os, 'access', return_value=False, autospec=True)
-    @mock.patch.object(os, 'stat', autospec=True)
-    @mock.patch.object(image_service.FileImageService, 'validate_href',
-                       autospec=True)
-    def test_download_copy_fail(self, _validate_mock, stat_mock, access_mock,
-                                open_mock, size_mock, copy_mock):
-        _validate_mock.return_value = self.href_path
-        stat_mock.return_value.st_dev = 'dev1'
-        file_mock = mock.MagicMock(spec=io.BytesIO)
-        file_mock.name = 'file'
-        input_mock = mock.MagicMock(spec=io.BytesIO)
-        open_mock.return_value = input_mock
-        self.assertRaises(exception.ImageDownloadFailed,
-                          self.service.download, self.href, file_mock)
-        _validate_mock.assert_called_once_with(mock.ANY, self.href)
-        self.assertEqual(2, stat_mock.call_count)
-        access_mock.assert_called_once_with(self.href_path, os.R_OK | os.W_OK)
-        size_mock.assert_called_once_with(self.href_path)
+        link_mock.assert_called_once_with(self.href_path, 'file')
+        copy_mock.assert_called_once_with(self.href_path, 'file')
 
 
 class ServiceGetterTestCase(base.TestCase):
