@@ -51,8 +51,6 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
             'step': 'build_raid', 'priority': 0, 'interface': 'deploy'}
 
     def __do_node_clean_validate_fail(self, mock_validate, clean_steps=None):
-        # InvalidParameterValue should cause node to go to CLEANFAIL
-        mock_validate.side_effect = exception.InvalidParameterValue('error')
         tgt_prov_state = states.MANAGEABLE if clean_steps else states.AVAILABLE
         node = obj_utils.create_test_node(
             self.context, driver='fake-hardware',
@@ -68,26 +66,42 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
         self.assertIsNone(node.fault)
         mock_validate.assert_called_once_with(mock.ANY, mock.ANY)
 
+    def __do_node_clean_validate_fail_invalid(self, mock_validate,
+                                              clean_steps=None):
+        # InvalidParameterValue should cause node to go to CLEANFAIL
+        mock_validate.side_effect = exception.InvalidParameterValue('error')
+        self.__do_node_clean_validate_fail(mock_validate,
+                                           clean_steps=clean_steps)
+
     @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                 autospec=True)
     def test__do_node_clean_automated_power_validate_fail(self, mock_validate):
-        self.__do_node_clean_validate_fail(mock_validate)
+        self.__do_node_clean_validate_fail_invalid(mock_validate)
 
     @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
                 autospec=True)
     def test__do_node_clean_manual_power_validate_fail(self, mock_validate):
-        self.__do_node_clean_validate_fail(mock_validate, clean_steps=[])
+        self.__do_node_clean_validate_fail_invalid(mock_validate,
+                                                   clean_steps=[])
 
     @mock.patch('ironic.drivers.modules.network.flat.FlatNetwork.validate',
                 autospec=True)
     def test__do_node_clean_automated_network_validate_fail(self,
                                                             mock_validate):
-        self.__do_node_clean_validate_fail(mock_validate)
+        self.__do_node_clean_validate_fail_invalid(mock_validate)
 
     @mock.patch('ironic.drivers.modules.network.flat.FlatNetwork.validate',
                 autospec=True)
     def test__do_node_clean_manual_network_validate_fail(self, mock_validate):
-        self.__do_node_clean_validate_fail(mock_validate, clean_steps=[])
+        self.__do_node_clean_validate_fail_invalid(mock_validate,
+                                                   clean_steps=[])
+
+    @mock.patch('ironic.drivers.modules.network.flat.FlatNetwork.validate',
+                autospec=True)
+    def test__do_node_clean_network_error_fail(self, mock_validate):
+        # NetworkError should cause node to go to CLEANFAIL
+        mock_validate.side_effect = exception.NetworkError()
+        self.__do_node_clean_validate_fail(mock_validate)
 
     @mock.patch.object(conductor_utils, 'LOG', autospec=True)
     @mock.patch.object(conductor_steps, 'set_node_cleaning_steps',
@@ -422,6 +436,36 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
         self.assertFalse(node.maintenance)
         self.assertIsNone(node.fault)
 
+    @mock.patch('ironic.drivers.modules.fake.FakePower.set_power_state',
+                autospec=True)
+    @mock.patch.object(n_flat.FlatNetwork, 'validate', autospec=True)
+    @mock.patch.object(conductor_steps, 'set_node_cleaning_steps',
+                       autospec=True)
+    def test_do_node_clean_steps_fail_poweroff(self, mock_steps, mock_validate,
+                                               mock_power, clean_steps=None,
+                                               invalid_exc=True):
+        if invalid_exc:
+            mock_steps.side_effect = exception.InvalidParameterValue('invalid')
+        else:
+            mock_steps.side_effect = exception.NodeCleaningFailure('failure')
+        tgt_prov_state = states.MANAGEABLE if clean_steps else states.AVAILABLE
+        self.config(poweroff_in_cleanfail=True, group='conductor')
+        node = obj_utils.create_test_node(
+            self.context, driver='fake-hardware',
+            uuid=uuidutils.generate_uuid(),
+            provision_state=states.CLEANING,
+            power_state=states.POWER_ON,
+            target_provision_state=tgt_prov_state)
+        with task_manager.acquire(
+                self.context, node.uuid, shared=False) as task:
+            cleaning.do_node_clean(task, clean_steps=clean_steps)
+            mock_validate.assert_called_once_with(mock.ANY, task)
+        node.refresh()
+        self.assertEqual(states.CLEANFAIL, node.provision_state)
+        self.assertEqual(tgt_prov_state, node.target_provision_state)
+        mock_steps.assert_called_once_with(mock.ANY, disable_ramdisk=False)
+        self.assertTrue(mock_power.called)
+
     def test__do_node_clean_automated_steps_fail(self):
         for invalid in (True, False):
             self.__do_node_clean_steps_fail(invalid_exc=invalid)
@@ -556,7 +600,8 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
             target_provision_state=tgt_prov_state,
             last_error=None,
             driver_internal_info={'clean_steps': self.clean_steps,
-                                  'clean_step_index': 0},
+                                  'clean_step_index': 0,
+                                  'cleaning_polling': True},
             clean_step=self.clean_steps[0])
         mock_execute.return_value = return_state
 
@@ -572,6 +617,7 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
         self.assertEqual(1, node.driver_internal_info['clean_step_index'])
         mock_execute.assert_called_once_with(
             mock.ANY, mock.ANY, self.clean_steps[1])
+        self.assertNotIn('cleaning_polling', node.driver_internal_info)
 
     def test_do_next_clean_step_continue_from_last_cleaning(self):
         self._do_next_clean_step_continue_from_last_cleaning(states.CLEANWAIT)
@@ -1092,6 +1138,36 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
     def test__do_next_clean_step_manual_bad_step_return_value(self):
         self._do_next_clean_step_bad_step_return_value(manual=True)
 
+    def _test_do_next_clean_step_handles_hold(self, start_state):
+        node = obj_utils.create_test_node(
+            self.context, driver='fake-hardware',
+            provision_state=start_state,
+            driver_internal_info={
+                'clean_steps': [
+                    {
+                        'step': 'hold',
+                        'priority': 10,
+                        'interface': 'power'
+                    }
+                ],
+                'clean_step_index': None},
+            clean_step=None)
+
+        with task_manager.acquire(
+                self.context, node.uuid, shared=False) as task:
+            cleaning.do_next_clean_step(task, 0)
+        node.refresh()
+
+        self.assertEqual(states.CLEANHOLD, node.provision_state)
+
+    def test_do_next_clean_step_handles_hold_from_active(self):
+        # Start is from the conductor
+        self._test_do_next_clean_step_handles_hold(states.CLEANING)
+
+    def test_do_next_clean_step_handles_hold_from_wait(self):
+        # Start is the continuation from a heartbeat.
+        self._test_do_next_clean_step_handles_hold(states.CLEANWAIT)
+
     @mock.patch.object(cleaning, 'do_next_clean_step', autospec=True)
     def _continue_node_clean(self, mock_next_step, skip=True):
         # test that skipping current step mechanism works
@@ -1124,7 +1200,8 @@ class DoNodeCleanTestCase(db_base.DbTestCase):
 
 class DoNodeCleanAbortTestCase(db_base.DbTestCase):
     @mock.patch.object(fake.FakeDeploy, 'tear_down_cleaning', autospec=True)
-    def _test_do_node_clean_abort(self, clean_step, tear_mock):
+    def _test_do_node_clean_abort(self, clean_step,
+                                  tear_mock=None):
         node = obj_utils.create_test_node(
             self.context, driver='fake-hardware',
             provision_state=states.CLEANWAIT,
@@ -1183,3 +1260,203 @@ class DoNodeCleanAbortTestCase(db_base.DbTestCase):
             self.assertIsNotNone(task.node.maintenance_reason)
             self.assertTrue(task.node.maintenance)
             self.assertEqual('clean failure', task.node.fault)
+
+    @mock.patch.object(fake.FakeDeploy, 'tear_down_cleaning', autospec=True)
+    def test__do_node_cleanhold_abort_tear_down_fail(self, tear_mock):
+        tear_mock.side_effect = Exception('Surprise')
+
+        node = obj_utils.create_test_node(
+            self.context, driver='fake-hardware',
+            provision_state=states.CLEANHOLD,
+            target_provision_state=states.MANAGEABLE,
+            clean_step={'step': 'hold', 'abortable': True})
+
+        with task_manager.acquire(self.context, node.uuid) as task:
+            cleaning.do_node_clean_abort(task)
+            tear_mock.assert_called_once_with(task.driver.deploy, task)
+            self.assertIsNotNone(task.node.last_error)
+            self.assertIsNotNone(task.node.maintenance_reason)
+            self.assertTrue(task.node.maintenance)
+            self.assertEqual('clean failure', task.node.fault)
+
+
+class DoNodeCleanTestChildNodes(db_base.DbTestCase):
+    def setUp(self):
+        super(DoNodeCleanTestChildNodes, self).setUp()
+        self.config(automated_clean=True, group='conductor')
+        self.power_off_parent = {
+            'step': 'power_off', 'priority': 4, 'interface': 'power'}
+        self.power_on_children = {
+            'step': 'power_on', 'priority': 5, 'interface': 'power',
+            'execute_on_child_nodes': True}
+        self.update_firmware_on_children = {
+            'step': 'update_firmware', 'priority': 10,
+            'interface': 'management', 'execute_on_child_nodes': True}
+        self.reboot_children = {
+            'step': 'reboot', 'priority': 5, 'interface': 'power',
+            'execute_on_child_nodes': True}
+        self.power_on_parent = {
+            'step': 'power_on', 'priority': 15, 'interface': 'power'}
+        self.clean_steps = [
+            self.power_off_parent,
+            self.power_on_children,
+            self.update_firmware_on_children,
+            self.reboot_children,
+            self.power_on_parent]
+        self.node = obj_utils.create_test_node(
+            self.context, driver='fake-hardware',
+            provision_state=states.CLEANING,
+            target_provision_state=states.MANAGEABLE,
+            last_error=None,
+            power_state=states.POWER_ON,
+            driver_internal_info={'agent_secret_token': 'old',
+                                  'clean_steps': self.clean_steps})
+
+    @mock.patch('ironic.drivers.modules.fake.FakePower.reboot',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakePower.set_power_state',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.network.flat.FlatNetwork.validate',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakePower.execute_clean_step',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakeManagement.'
+                'execute_clean_step', autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakeDeploy.execute_clean_step',
+                autospec=True)
+    def test_do_next_clean_step_with_children(
+            self, mock_deploy, mock_mgmt, mock_power, mock_pv, mock_nv,
+            mock_sps, mock_reboot):
+        child_node1 = obj_utils.create_test_node(
+            self.context,
+            uuid=uuidutils.generate_uuid(),
+            driver='fake-hardware',
+            last_error=None,
+            power_state=states.POWER_OFF,
+            parent_node=self.node.uuid)
+        child_node2 = obj_utils.create_test_node(
+            self.context,
+            uuid=uuidutils.generate_uuid(),
+            driver='fake-hardware',
+            last_error=None,
+            power_state=states.POWER_OFF,
+            parent_node=self.node.uuid)
+
+        mock_deploy.return_value = None
+        mock_mgmt.return_value = None
+        mock_power.return_value = None
+        child1_updated_at = str(child_node1.updated_at)
+        child2_updated_at = str(child_node2.updated_at)
+        with task_manager.acquire(
+                self.context, self.node.uuid, shared=False) as task:
+
+            cleaning.do_next_clean_step(task, 0,
+                                        disable_ramdisk=True)
+        self.node.refresh()
+        child_node1.refresh()
+        child_node2.refresh()
+
+        # Confirm the objects *did* receive locks.
+        self.assertNotEqual(child1_updated_at, child_node1.updated_at)
+        self.assertNotEqual(child2_updated_at, child_node2.updated_at)
+
+        # Confirm the child nodes have no errors
+        self.assertFalse(child_node1.maintenance)
+        self.assertFalse(child_node2.maintenance)
+        self.assertIsNone(child_node1.last_error)
+        self.assertIsNone(child_node2.last_error)
+        self.assertIsNone(self.node.last_error)
+
+        # Confirm the call counts expected
+        self.assertEqual(0, mock_deploy.call_count)
+        self.assertEqual(2, mock_mgmt.call_count)
+        self.assertEqual(0, mock_power.call_count)
+        self.assertEqual(0, mock_nv.call_count)
+        self.assertEqual(0, mock_pv.call_count)
+        self.assertEqual(4, mock_sps.call_count)
+        self.assertEqual(2, mock_reboot.call_count)
+        mock_sps.assert_has_calls([
+            mock.call(mock.ANY, mock.ANY, 'power off', timeout=None),
+            mock.call(mock.ANY, mock.ANY, 'power on', timeout=None),
+            mock.call(mock.ANY, mock.ANY, 'power on', timeout=None)])
+
+    @mock.patch('ironic.drivers.modules.fake.FakePower.set_power_state',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.network.flat.FlatNetwork.validate',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakePower.validate',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakePower.execute_clean_step',
+                autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakeManagement.'
+                'execute_clean_step', autospec=True)
+    @mock.patch('ironic.drivers.modules.fake.FakeDeploy.execute_clean_step',
+                autospec=True)
+    def test_do_next_clean_step_with_children_by_uuid(
+            self, mock_deploy, mock_mgmt, mock_power, mock_pv, mock_nv,
+            mock_sps):
+        child_node1 = obj_utils.create_test_node(
+            self.context,
+            uuid=uuidutils.generate_uuid(),
+            driver='fake-hardware',
+            last_error=None,
+            parent_node=self.node.uuid)
+        child_node2 = obj_utils.create_test_node(
+            self.context,
+            uuid=uuidutils.generate_uuid(),
+            driver='fake-hardware',
+            last_error=None,
+            parent_node=self.node.uuid)
+        power_on_children = {
+            'step': 'power_on', 'priority': 5, 'interface': 'power',
+            'execute_on_child_nodes': True,
+            'limit_child_node_execution': [child_node1.uuid]}
+        update_firmware_on_children = {
+            'step': 'update_firmware', 'priority': 10,
+            'interface': 'management',
+            'execute_on_child_nodes': True,
+            'limit_child_node_execution': [child_node1.uuid]}
+        power_on_parent = {
+            'step': 'not_power', 'priority': 15, 'interface': 'power'}
+        clean_steps = [power_on_children, update_firmware_on_children,
+                       power_on_parent]
+        dii = self.node.driver_internal_info
+        dii['clean_steps'] = clean_steps
+        self.node.driver_internal_info = dii
+        self.node.save()
+
+        mock_deploy.return_value = None
+        mock_mgmt.return_value = None
+        mock_power.return_value = None
+        child1_updated_at = str(child_node1.updated_at)
+
+        with task_manager.acquire(
+                self.context, self.node.uuid, shared=False) as task:
+
+            cleaning.do_next_clean_step(task, 0,
+                                        disable_ramdisk=True)
+        self.node.refresh()
+        child_node1.refresh()
+        child_node2.refresh()
+
+        # Confirm the objects *did* receive locks.
+        self.assertNotEqual(child1_updated_at, child_node1.updated_at)
+        self.assertIsNone(child_node2.updated_at)
+
+        # Confirm the child nodes have no errors
+        self.assertFalse(child_node1.maintenance)
+        self.assertFalse(child_node2.maintenance)
+        self.assertIsNone(child_node1.last_error)
+        self.assertIsNone(child_node2.last_error)
+        self.assertIsNone(self.node.last_error)
+
+        # Confirm the call counts expected
+        self.assertEqual(0, mock_deploy.call_count)
+        self.assertEqual(1, mock_mgmt.call_count)
+        self.assertEqual(1, mock_power.call_count)
+        self.assertEqual(0, mock_nv.call_count)
+        self.assertEqual(0, mock_pv.call_count)
+        mock_sps.assert_has_calls([
+            mock.call(mock.ANY, mock.ANY, 'power on', timeout=None)])
