@@ -16,8 +16,8 @@
 import datetime
 from unittest import mock
 
-from oslo_utils import importutils
 from oslo_utils import units
+import sushy
 
 from ironic.common import boot_devices
 from ironic.common import boot_modes
@@ -39,8 +39,6 @@ from ironic.tests.unit.db import base as db_base
 from ironic.tests.unit.db import utils as db_utils
 from ironic.tests.unit.objects import utils as obj_utils
 
-sushy = importutils.try_import('sushy')
-
 INFO_DICT = db_utils.get_test_redfish_info()
 
 
@@ -59,7 +57,6 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
         self.system_uuid = 'ZZZ--XXX-YYY'
         self.chassis_uuid = 'XXX-YYY-ZZZ'
-        self.drive_uuid = 'ZZZ-YYY-XXX'
 
     def init_system_mock(self, system_mock, **properties):
 
@@ -83,13 +80,6 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
         return system_mock
 
-    @mock.patch.object(redfish_mgmt, 'sushy', None)
-    def test_loading_error(self):
-        self.assertRaisesRegex(
-            exception.DriverLoadError,
-            'Unable to import the sushy library',
-            redfish_mgmt.RedfishManagement)
-
     def test_get_properties(self):
         with task_manager.acquire(self.context, self.node.uuid,
                                   shared=True) as task:
@@ -122,7 +112,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 (boot_devices.PXE, sushy.BOOT_SOURCE_TARGET_PXE),
                 (boot_devices.DISK, sushy.BOOT_SOURCE_TARGET_HDD),
                 (boot_devices.CDROM, sushy.BOOT_SOURCE_TARGET_CD),
-                (boot_devices.BIOS, sushy.BOOT_SOURCE_TARGET_BIOS_SETUP)
+                (boot_devices.BIOS, sushy.BOOT_SOURCE_TARGET_BIOS_SETUP),
+                (boot_devices.UEFIHTTP, sushy.BOOT_SOURCE_TARGET_UEFI_HTTP)
             ]
 
             for target, expected in expected_values:
@@ -131,7 +122,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 # Asserts
                 fake_system.set_system_boot_options.assert_has_calls(
                     [mock.call(expected,
-                               enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)])
+                               enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                               http_boot_uri=None)])
                 mock_get_system.assert_called_with(task.node)
                 self.assertNotIn('redfish_boot_device',
                                  task.node.driver_internal_info)
@@ -157,7 +149,7 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
                 fake_system.set_system_boot_options.assert_has_calls(
                     [mock.call(sushy.BOOT_SOURCE_TARGET_PXE,
-                               enabled=expected)])
+                               enabled=expected, http_boot_uri=None)])
                 mock_get_system.assert_called_with(task.node)
                 self.assertNotIn('redfish_boot_device',
                                  task.node.driver_internal_info)
@@ -184,7 +176,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                     task, boot_devices.PXE, persistent=target)
 
                 fake_system.set_system_boot_options.assert_has_calls(
-                    [mock.call(sushy.BOOT_SOURCE_TARGET_PXE, enabled=None)])
+                    [mock.call(sushy.BOOT_SOURCE_TARGET_PXE, enabled=None,
+                               http_boot_uri=None)])
                 mock_get_system.assert_called_with(task.node)
 
                 # Reset mocks
@@ -206,7 +199,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 task.driver.management.set_boot_device, task, boot_devices.PXE)
             fake_system.set_system_boot_options.assert_called_once_with(
                 sushy.BOOT_SOURCE_TARGET_PXE,
-                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)
+                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                http_boot_uri=None)
             mock_get_system.assert_called_once_with(task.node)
             self.assertNotIn('redfish_boot_device',
                              task.node.driver_internal_info)
@@ -233,7 +227,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                     task.driver.management.set_boot_device, task,
                     boot_devices.PXE, persistent=target)
                 fake_system.set_system_boot_options.assert_called_once_with(
-                    sushy.BOOT_SOURCE_TARGET_PXE, enabled=None)
+                    sushy.BOOT_SOURCE_TARGET_PXE, enabled=None,
+                    http_boot_uri=None)
                 mock_get_system.assert_called_once_with(task.node)
                 self.assertNotIn('redfish_boot_device',
                                  task.node.driver_internal_info)
@@ -259,9 +254,11 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 task, boot_devices.PXE, persistent=True)
             fake_system.set_system_boot_options.assert_has_calls([
                 mock.call(sushy.BOOT_SOURCE_TARGET_PXE,
-                          enabled=sushy.BOOT_SOURCE_ENABLED_CONTINUOUS),
+                          enabled=sushy.BOOT_SOURCE_ENABLED_CONTINUOUS,
+                          http_boot_uri=None),
                 mock.call(sushy.BOOT_SOURCE_TARGET_PXE,
-                          enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)
+                          enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                          http_boot_uri=None)
             ])
             mock_get_system.assert_called_with(task.node)
 
@@ -293,7 +290,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 task.driver.management.set_boot_device(
                     task, boot_devices.PXE, persistent=True)
                 fake_system.set_system_boot_options.assert_called_once_with(
-                    sushy.BOOT_SOURCE_TARGET_PXE, enabled=expected)
+                    sushy.BOOT_SOURCE_TARGET_PXE, enabled=expected,
+                    http_boot_uri=None)
                 if vendor == 'SuperMicro':
                     mock_sync_boot_mode.assert_called_once_with(task)
                 else:
@@ -304,6 +302,28 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                 mock_sync_boot_mode.reset_mock()
                 mock_get_system.reset_mock()
 
+    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
+    def test_set_boot_device_http_boot(self, mock_get_system):
+        fake_system = mock.Mock()
+        mock_get_system.return_value = fake_system
+        self.node.driver_internal_info = {
+            'redfish_uefi_http_url': 'http://foo.url'}
+        self.node.save()
+        with task_manager.acquire(self.context, self.node.uuid,
+                                  shared=False) as task:
+            task.driver.management.set_boot_device(task,
+                                                   boot_devices.UEFIHTTP)
+            fake_system.set_system_boot_options.assert_has_calls(
+                [mock.call(sushy.BOOT_SOURCE_TARGET_UEFI_HTTP,
+                           enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                           http_boot_uri='http://foo.url')])
+            mock_get_system.assert_called_with(task.node)
+            self.assertNotIn('redfish_boot_device',
+                             task.node.driver_internal_info)
+            task.node.refresh()
+            self.assertNotIn('redfish_uefi_http_url',
+                             task.node.driver_internal_info)
+
     def test_restore_boot_device(self):
         fake_system = mock.Mock()
         with task_manager.acquire(self.context, self.node.uuid,
@@ -316,7 +336,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
             fake_system.set_system_boot_options.assert_called_once_with(
                 sushy.BOOT_SOURCE_TARGET_HDD,
-                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)
+                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                http_boot_uri=None)
             # The stored boot device is kept intact
             self.assertEqual(
                 boot_devices.DISK,
@@ -333,7 +354,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
             fake_system.set_system_boot_options.assert_called_once_with(
                 sushy.BOOT_SOURCE_TARGET_HDD,
-                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)
+                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                http_boot_uri=None)
             # The stored boot device is kept intact
             self.assertEqual(
                 "hdd",
@@ -363,7 +385,8 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
             fake_system.set_system_boot_options.assert_called_once_with(
                 sushy.BOOT_SOURCE_TARGET_HDD,
-                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE)
+                enabled=sushy.BOOT_SOURCE_ENABLED_ONCE,
+                http_boot_uri=None)
             self.assertTrue(mock_log.called)
             # The stored boot device is kept intact
             self.assertEqual(
@@ -670,7 +693,7 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
         self.assertEqual(expected, sensors)
 
-    def test__get_sensors_data_drive(self):
+    def test__get_sensors_data_drive_simple_storage(self):
         attributes = {
             'name': '32ADF365C6C1B7BD',
             'manufacturer': 'IBM',
@@ -682,14 +705,58 @@ class RedfishManagementTestCase(db_base.DbTestCase):
             }
         }
 
-        mock_system = mock.MagicMock(identity='ZZZ-YYY-XXX')
+        mock_system = mock.MagicMock(spec=sushy.resources.system.system.System)
+        mock_system.identity = 'ZZZ-YYY-XXX'
         mock_drive = mock.MagicMock(**attributes)
         mock_drive.name = attributes['name']
         mock_drive.status = mock.MagicMock(**attributes['status'])
-        mock_storage = mock.MagicMock()
-        mock_storage.devices = [mock_drive]
+        mock_storage = mock.MagicMock(
+            spec=sushy.resources.system.storage.storage.Storage)
+        mock_storage.drives = [mock_drive]
         mock_storage.identity = 'XXX-YYY-ZZZ'
-        mock_system.simple_storage.get_members.return_value = [mock_storage]
+        mock_system.storage.get_members.return_value = [mock_storage]
+        mock_system.simple_storage = {}
+
+        with task_manager.acquire(self.context, self.node.uuid,
+                                  shared=True) as task:
+            sensors = task.driver.management._get_sensors_drive(mock_system)
+
+        expected = {
+            '32ADF365C6C1B7BD:XXX-YYY-ZZZ@ZZZ-YYY-XXX': {
+                'capacity_bytes': 3750000000,
+                'health': 'OK',
+                'name': '32ADF365C6C1B7BD',
+                'model': 'IBM 350A',
+                'state': 'enabled'
+            }
+        }
+
+        self.assertEqual(expected, sensors)
+
+    def test__get_sensors_data_drive_storage(self):
+        attributes = {
+            'name': '32ADF365C6C1B7BD',
+            'manufacturer': 'IBM',
+            'model': 'IBM 350A',
+            'capacity_bytes': 3750000000,
+            'status': {
+                'health': 'OK',
+                'state': 'enabled'
+            }
+        }
+
+        mock_system = mock.MagicMock(spec=sushy.resources.system.system.System)
+        mock_system.identity = 'ZZZ-YYY-XXX'
+        mock_drive = mock.MagicMock(**attributes)
+        mock_drive.name = attributes['name']
+        mock_drive.status = mock.MagicMock(**attributes['status'])
+        mock_simple_storage = mock.MagicMock(
+            spec=sushy.resources.system.simple_storage.SimpleStorage)
+        mock_simple_storage.devices = [mock_drive]
+        mock_simple_storage.identity = 'XXX-YYY-ZZZ'
+        mock_system.simple_storage.get_members.return_value = [
+            mock_simple_storage]
+        mock_system.storage = {}
 
         with task_manager.acquire(self.context, self.node.uuid,
                                   shared=True) as task:
@@ -757,13 +824,17 @@ class RedfishManagementTestCase(db_base.DbTestCase):
             uuid=self.chassis_uuid,
             indicator_led=sushy.INDICATOR_LED_LIT)
         fake_drive = mock.Mock(
-            uuid=self.drive_uuid,
+            identity='drive1',
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_storage = mock.Mock(
+            identity='storage1',
+            drives=[fake_drive])
         fake_system = mock.Mock(
             uuid=self.system_uuid,
             chassis=[fake_chassis],
-            simple_storage=mock.MagicMock(drives=[fake_drive]),
+            storage=mock.MagicMock(),
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_system.storage.get_members.return_value = [fake_storage]
 
         mock_get_system.return_value = fake_system
 
@@ -795,7 +866,7 @@ class RedfishManagementTestCase(db_base.DbTestCase):
                     }
                 },
                 components.DISK: {
-                    'ZZZ-YYY-XXX': {
+                    'storage1:drive1': {
                         "readonly": False,
                         "states": [
                             indicator_states.BLINKING,
@@ -814,13 +885,17 @@ class RedfishManagementTestCase(db_base.DbTestCase):
             uuid=self.chassis_uuid,
             indicator_led=sushy.INDICATOR_LED_LIT)
         fake_drive = mock.Mock(
-            uuid=self.drive_uuid,
+            identity='drive1',
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_storage = mock.Mock(
+            identity='storage1',
+            drives=[fake_drive])
         fake_system = mock.Mock(
             uuid=self.system_uuid,
             chassis=[fake_chassis],
-            simple_storage=mock.MagicMock(drives=[fake_drive]),
+            storage=mock.MagicMock(),
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_system.storage.get_members.return_value = [fake_storage]
 
         mock_get_system.return_value = fake_system
 
@@ -834,19 +909,53 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
             mock_get_system.assert_called_once_with(task.node)
 
+    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
+    def test_set_indicator_state_disk(self, mock_get_system):
+        fake_chassis = mock.Mock(
+            uuid=self.chassis_uuid,
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_drive = mock.Mock(
+            identity='drive1',
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_storage = mock.Mock(
+            identity='storage1',
+            drives=[fake_drive])
+        fake_system = mock.Mock(
+            uuid=self.system_uuid,
+            chassis=[fake_chassis],
+            storage=mock.MagicMock(),
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_system.storage.get_members.return_value = [fake_storage]
+
+        mock_get_system.return_value = fake_system
+
+        with task_manager.acquire(self.context, self.node.uuid,
+                                  shared=False) as task:
+            task.driver.management.set_indicator_state(
+                task, components.DISK, 'storage1:drive1', indicator_states.ON)
+
+            fake_drive.set_indicator_led.assert_called_once_with(
+                sushy.INDICATOR_LED_LIT)
+
+            mock_get_system.assert_called_once_with(task.node)
+
     @mock.patch.object(redfish_utils, 'get_system', autospec=True)
     def test_get_indicator_state(self, mock_get_system):
         fake_chassis = mock.Mock(
             uuid=self.chassis_uuid,
             indicator_led=sushy.INDICATOR_LED_LIT)
         fake_drive = mock.Mock(
-            uuid=self.drive_uuid,
+            identity='drive1',
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_storage = mock.Mock(
+            identity='storage1',
+            drives=[fake_drive])
         fake_system = mock.Mock(
             uuid=self.system_uuid,
             chassis=[fake_chassis],
-            simple_storage=mock.MagicMock(drives=[fake_drive]),
+            storage=mock.MagicMock(),
             indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_system.storage.get_members.return_value = [fake_storage]
 
         mock_get_system.return_value = fake_system
 
@@ -860,6 +969,36 @@ class RedfishManagementTestCase(db_base.DbTestCase):
 
             self.assertEqual(indicator_states.ON, state)
 
+    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
+    def test_get_indicator_state_disk(self, mock_get_system):
+        fake_chassis = mock.Mock(
+            uuid=self.chassis_uuid,
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_drive = mock.Mock(
+            identity='drive1',
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_storage = mock.Mock(
+            identity='storage1',
+            drives=[fake_drive])
+        fake_system = mock.Mock(
+            uuid=self.system_uuid,
+            chassis=[fake_chassis],
+            storage=mock.MagicMock(),
+            indicator_led=sushy.INDICATOR_LED_LIT)
+        fake_system.storage.get_members.return_value = [fake_storage]
+
+        mock_get_system.return_value = fake_system
+
+        with task_manager.acquire(self.context, self.node.uuid,
+                                  shared=False) as task:
+
+            state = task.driver.management.get_indicator_state(
+                task, components.DISK, 'storage1:drive1')
+
+            mock_get_system.assert_called_once_with(task.node)
+
+            self.assertEqual(indicator_states.ON, state)
+
     @mock.patch.object(redfish_utils, 'get_system', autospec=True)
     def test_detect_vendor(self, mock_get_system):
         mock_get_system.return_value.manufacturer = "Fake GmbH"
@@ -1740,3 +1879,24 @@ class RedfishManagementTestCase(db_base.DbTestCase):
         with task_manager.acquire(self.context, self.node.uuid,
                                   shared=True) as task:
             self.assertIsNone(task.driver.management.get_mac_addresses(task))
+
+    @mock.patch.object(redfish_boot, 'insert_vmedia', autospec=True)
+    def test_attach_virtual_media(self, mock_insert_vmedia):
+        with task_manager.acquire(self.context, self.node.uuid) as task:
+            task.driver.management.attach_virtual_media(task, 'cdrom',
+                                                        'http://test.iso')
+            mock_insert_vmedia.assert_called_once_with(task, 'http://test.iso',
+                                                       sushy.VIRTUAL_MEDIA_CD)
+
+    @mock.patch.object(redfish_boot, 'eject_vmedia', autospec=True)
+    def test_detach_virtual_media(self, mock_eject_vmedia):
+        with task_manager.acquire(self.context, self.node.uuid) as task:
+            task.driver.management.detach_virtual_media(task, ['cdrom'])
+            mock_eject_vmedia.assert_called_once_with(task,
+                                                      sushy.VIRTUAL_MEDIA_CD)
+
+    @mock.patch.object(redfish_boot, 'eject_vmedia', autospec=True)
+    def test_detach_virtual_media_all(self, mock_eject_vmedia):
+        with task_manager.acquire(self.context, self.node.uuid) as task:
+            task.driver.management.detach_virtual_media(task)
+            mock_eject_vmedia.assert_called_once_with(task)
